from time import sleep
import requests
import numpy as np
from .exceptions import DatameshConnectError
import os


# Timeouts in seconds to establish connection to datamesh services
# for read types of operations
DATAMESH_CONNECT_TIMEOUT = os.getenv("DATAMESH_CONNECT_TIMEOUT", 3.05)
DATAMESH_CONNECT_TIMEOUT = (
    None if DATAMESH_CONNECT_TIMEOUT == "None" else float(DATAMESH_CONNECT_TIMEOUT)
)

# Timeout in seconds to read data from datamesh services
# for small json payloads type of operations
DATAMESH_READ_TIMEOUT = os.getenv("DATAMESH_READ_TIMEOUT", 10)
DATAMESH_READ_TIMEOUT = (
    None if DATAMESH_READ_TIMEOUT == "None" else float(DATAMESH_READ_TIMEOUT)
)

# Timeout in seconds for staging endpoint
DATAMESH_STAGE_READ_TIMEOUT = os.getenv("DATAMESH_STAGE_READ_TIMEOUT", 900)
DATAMESH_STAGE_READ_TIMEOUT = (
    None if DATAMESH_STAGE_READ_TIMEOUT == "None" else float(DATAMESH_STAGE_READ_TIMEOUT)
)

# Timeout in seconds for bulk download operations
DATAMESH_DOWNLOAD_TIMEOUT = os.getenv("DATAMESH_DOWNLOAD_TIMEOUT", 900)
DATAMESH_DOWNLOAD_TIMEOUT = (
    None if DATAMESH_DOWNLOAD_TIMEOUT == "None" else float(DATAMESH_DOWNLOAD_TIMEOUT)
)

# Timeout in seconds for bulk write operations
DATAMESH_WRITE_TIMEOUT = os.getenv("DATAMESH_WRITE_TIMEOUT", "None")
DATAMESH_WRITE_TIMEOUT = (
    None if DATAMESH_WRITE_TIMEOUT == "None" else float(DATAMESH_WRITE_TIMEOUT)
)

# Timeout in seconds for zarr chunk read operations
DATAMESH_CHUNK_READ_TIMEOUT = os.getenv("DATAMESH_CHUNK_READ_TIMEOUT", 60)
DATAMESH_CHUNK_READ_TIMEOUT = (
    None if DATAMESH_CHUNK_READ_TIMEOUT == "None" else float(DATAMESH_CHUNK_READ_TIMEOUT)
)

# Timeout in seconds for zarr chunk write operations
# much larger than for read seems to be required possibly because write acknowledgement
# occurs after the chunk has been fully written
DATAMESH_CHUNK_WRITE_TIMEOUT = os.getenv("DATAMESH_CHUNK_WRITE_TIMEOUT", 600)
DATAMESH_CHUNK_WRITE_TIMEOUT = (
    None if DATAMESH_CHUNK_WRITE_TIMEOUT == "None" else float(DATAMESH_CHUNK_WRITE_TIMEOUT)
)


def retried_request(
    url,
    method="GET",
    data=None,
    params=None,
    headers=None,
    retries=8,
    timeout=(DATAMESH_CONNECT_TIMEOUT, DATAMESH_READ_TIMEOUT),
    verify=True,
):
    """
    Retried request function with exponential backoff

    Parameters
    ----------
    url : str
        URL to request
    method : str, optional
        HTTP method, by default "GET"
    data : str, optional
        Request data, by default None
    headers : dict, optional
        Request headers, by default None
    retries : int, optional
        Number of retries, by default 8
    timeout : tupe(float, float), optional
        Request connect and read timeout in seconds, by default (3.05, 10)

    Returns
    -------
    requests.Response
        Response object

    Raises
    ------
    requests.RequestException
        If request fails

    """
    retried = 0
    while retried < retries:
        try:
            resp = requests.request(
                method=method,
                url=url,
                data=data,
                params=params,
                headers=headers,
                timeout=timeout,
                verify=verify,
            )
            # Bad Gateway results in waiting for 10 seconds
            # and retrying
            if resp.status_code == 502:
                sleep(30)
                raise requests.RequestException
        except (
            requests.RequestException,
            requests.ReadTimeout,
            requests.ConnectionError,
            requests.ConnectTimeout,
        ) as e:
            sleep(0.1 * 2**retried)
            retried += 1
            if retried == retries:
                raise DatameshConnectError(
                    f"Failed to connect to {url} after {retries} retries with error: {e}"
                )
        else:
            return resp



"""  Datamesh CLI commands """

import click

from . import main
from .utils import get_bearer_token
from .auth import login_required
from ..datamesh.connection import Connector


@main.group(help='Datamesh Datasources commands')
def datamesh():
    pass

@datamesh.group(name='list')
def list_():
    pass

class DatameshClient:
    def __init__(self, ctx: click.Context) -> None:
        service_url = f'https://datamesh.{ctx.obj.domain}/'
        self.connector = Connector(
            token=get_bearer_token(ctx),
            service=service_url
        )

    def __enter__(self):
        return self.connector

    def __exit__(self, exc_type, exc_value, traceback):
        pass

@list_.command()
@click.pass_context
@click.option('--search', help='Search string', default=None, type=str)
@click.option('--limit', help='Limit results', default=10)
@login_required
def datasources(ctx: click.Context, search: str, limit: int):
    with DatameshClient(ctx) as client:
        datasources = client.get_catalog(search=search, limit=limit)
        click.echo(datasources)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `oceanum` package."""
from pathlib import Path
import json
import os
import pytest
import shapely
import xarray as xr
from uuid import uuid4
from click.testing import CliRunner

from oceanum.datamesh import Connector, Datasource
from oceanum.datamesh.query import GeoFilter, TimeFilter
from oceanum.datamesh.utils import retried_request
from oceanum import cli


HERE = Path(__file__).parent


@pytest.fixture
def conn():
    """Connection fixture"""
    return Connector(os.environ["DATAMESH_TOKEN"])


@pytest.fixture
def dataset():
    ds = xr.open_dataset(HERE / "data/grid_data_1.nc")
    return ds.isel(longitude=slice(0, 2), latitude=slice(0, 2), time=slice(0, 2))


def test_catalog_ignore_broken(conn, dataset):
    tag = str(uuid4())[:8]
    datasource_id = f"test-dataset-catalog-{tag}"
    try:
        x0, x1 = dataset.longitude.values[[0, -1]]
        y0, y1 = dataset.latitude.values[[0, -1]]
        conn.write_datasource(
            datasource_id,
            dataset,
            overwrite=True,
            geom=shapely.geometry.box(x0, y0, x1, y1),
            tags=[tag, "test", "dataset", "catalog"],
        )
        # Corrupt the metadata
        ds = conn.get_datasource(datasource_id)
        data = json.loads(
            ds.model_dump_json(by_alias=True, warnings=False).encode("utf-8", "ignore")
        )
        data["coordinates"] = {"dummy_key": "dummy_val"}
        data = json.dumps(data).encode("utf-8", "ignore")
        headers = {**conn._auth_headers, "Content-Type": "application/json"}
        resp = retried_request(
            f"{conn._proto}://{conn._host}/datasource/{datasource_id}/",
            method="PATCH",
            data=data,
            headers=headers,
        )
        # The catalog should return None for this datasource
        cat = conn.get_catalog(search=datasource_id)
        assert None in list(cat)
    finally:
        # Need to put it right before delete otherwise the gateway part
        # of the delete won't succeed as it won't be able to parse the metadata
        # due to it failing pydantic checks
        data = json.loads(
            ds.model_dump_json(by_alias=True, warnings=False).encode("utf-8", "ignore")
        )
        data["coordinates"] = {}
        data = json.dumps(data).encode("utf-8", "ignore")
        resp = retried_request(
            f"{conn._proto}://{conn._host}/datasource/{datasource_id}/",
            method="PATCH",
            data=data,
            headers=headers,
        )
        conn.delete_datasource(datasource_id)


def test_catalog_search(conn):
    cat = conn.get_catalog(search="wave")
    ds0 = cat.ids[0]
    assert ds0 in str(cat)
    assert isinstance(cat[ds0], Datasource)
    assert len(cat)


def test_catalog_timefilter(conn):
    cat = conn.get_catalog(timefilter=TimeFilter(times=["2010-01-01", "2020-01-01"]))
    ds0 = cat.ids[0]
    assert ds0 in str(cat)
    assert isinstance(cat[ds0], Datasource)
    assert len(cat)


def test_catalog_timefilter_none(conn):
    cat = conn.get_catalog(timefilter=["2010-01-01", None])
    ds0 = cat.ids[0]
    assert ds0 in str(cat)
    assert isinstance(cat[ds0], Datasource)
    assert len(cat)


def test_catalog_geofilter_shapely(conn):
    bbox = shapely.geometry.box(0, 0, 10, 10)
    cat = conn.get_catalog(geofilter=bbox)
    ds0 = cat.ids[0]
    assert ds0 in str(cat)
    assert isinstance(cat[ds0], Datasource)
    assert len(cat)


def test_catalog_geofilter_bbox(conn):
    geofilter = GeoFilter(type="bbox", geom=[0, 0, 10, 10])
    cat = conn.get_catalog(geofilter=geofilter)
    ds0 = cat.ids[0]
    assert ds0 in str(cat)
    assert isinstance(cat[ds0], Datasource)
    assert len(cat)


def test_catalog_geofilter_feature(conn):
    geofilter = GeoFilter(
        type="feature",
        geom={
            "type": "Feature",
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]]],
            },
        },
    )
    cat = conn.get_catalog(geofilter=geofilter)
    ds0 = cat.ids[0]
    assert ds0 in str(cat)
    assert isinstance(cat[ds0], Datasource)
    assert len(cat)



import time
import urllib.parse

import click
import requests

from functools import update_wrapper

from . import main
from .symbols import spin, err
from .models import DeviceCodeResponse, TokenResponse, Auth0Config


class Auth0Client:
    def __init__(self, ctx: click.Context) -> None:
        self.config = ctx.obj.auth0
        self.ctx = ctx

    def _request(self, method, endpoint, **kwargs) -> requests.Response:
        url = f'https://{self.config.domain}/' + endpoint
        kwargs.setdefault('headers', {
            'Content-Type': 'application/x-www-form-urlencoded'
        })
        response = requests.request(method, url, **kwargs)
        response.raise_for_status()
        return response

    def get_device_code(self) -> DeviceCodeResponse:
        data = {
            'client_id': self.config.client_id,
            'scope': 'offline_access',
        }
        response = self._request('POST', 'oauth/device/code', data=data)
        device_code = DeviceCodeResponse(**response.json())
        return device_code
    
    def get_token(self, device_code: str) -> TokenResponse:
        data = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
            'client_id': self.config.client_id,
            'device_code': device_code
        }
        response = self._request('POST', 'oauth/token', data=data)
        return TokenResponse(domain=self.ctx.obj.domain, **response.json())
    
    def refresh_token(self, token: TokenResponse) -> TokenResponse:
        if token.refresh_token is None:
            raise Exception('Token does not have a refresh token!')
        data = {
            'grant_type': 'refresh_token',
            'client_id': self.config.client_id,
            'refresh_token': token.refresh_token
        }
        response = self._request('POST', 'oauth/token', data=data)
        return TokenResponse(domain=self.ctx.obj.domain, **response.json())
    
    def wait_for_confirmation(self, device_code: DeviceCodeResponse) -> TokenResponse|None:
        t0 = time.time()
        token = None
        while time.time() - t0 < device_code.expires_in:
            try:
                time.sleep(device_code.interval)
                token = self.get_token(device_code.device_code)
            except requests.HTTPError as e:
                #print(f'Error fetching auth token: {e}')
                continue
            else:
                token.save()
                break
        return token


@main.group(help='Authentication commands')
def auth():
    pass

def login_required(func):
    @click.pass_context
    def refresh_wrapper(ctx: click.Context, *args, **kwargs):
        if not ctx.obj.token or not ctx.obj.token.refresh_token:
            raise Exception('You need to login first! Run `oceanum auth login`.')
        elif ctx.obj.token.is_expired and ctx.obj.token.refresh_token:
            click.echo(f' {spin} Refreshing access token...')
            try:
                token = Auth0Client(ctx=ctx).refresh_token(ctx.obj.token)
            except requests.HTTPError as e:
                if e.response.status_code == 403:
                    click.echo(f' {err} Failed to refresh token! Please log-in and try again...')
                    ctx.obj.token.delete()
                    return ctx.invoke(login)
                else:
                    raise e
            else:
                token.save()
                ctx.obj.token = token
        return ctx.invoke(func, *args, **kwargs)
    return update_wrapper(refresh_wrapper, func)

@auth.command()
@click.pass_context
def login(ctx: click.Context):
    click.echo('Logging in...')
    client = Auth0Client(ctx=ctx)
    device_code = client.get_device_code()
    click.echo(f'Please visit {device_code.verification_uri_complete} and confirm the code: {device_code.user_code}')
    client.wait_for_confirmation(device_code)
    click.echo('Logged in successfully!')

@auth.command()
@click.pass_context
@login_required
def logout(ctx: click.Context):
    ctx.obj.token.delete()
    click.echo('Logged out successfully!')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for verify parameter propagation in oceanum package."""
import os
import pytest
from unittest.mock import patch, MagicMock

from oceanum.datamesh import Connector
from oceanum.datamesh.zarr import ZarrClient


@pytest.fixture
def mock_request():
    """Mock the requests.request function to check verify parameter"""
    with patch("requests.request") as mock_req:
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"id": "test-id", "name": "test"}
        mock_req.return_value = mock_response
        yield mock_req


@pytest.fixture
def conn():
    """Connection fixture with verify=False"""
    return Connector(os.environ["DATAMESH_TOKEN"], verify=False)


def test_verify_parameter_default(mock_request):
    """Test that verify parameter defaults to True and is propagated to requests"""
    conn = Connector(token="test-token")

    # Test a method that makes a request
    conn._status()

    # Check that verify=True was passed to the request
    call_kwargs = mock_request.call_args.kwargs
    assert "verify" in call_kwargs
    assert call_kwargs["verify"] is True


def test_verify_parameter_false(mock_request):
    """Test that verify=False is properly propagated to requests"""
    conn = Connector(token="test-token", verify=False)

    # Test a method that makes a request
    conn._status()

    # Check that verify=False was passed to the request
    call_kwargs = mock_request.call_args.kwargs
    assert "verify" in call_kwargs
    assert call_kwargs["verify"] is False


def test_verify_parameter_in_metadata_request(mock_request):
    """Test that verify parameter is propagated to metadata requests"""
    conn = Connector(token="test-token", verify=False)

    # Test metadata request
    conn._metadata_request("test-id")

    # Check that verify=False was passed to the request
    call_kwargs = mock_request.call_args.kwargs
    assert "verify" in call_kwargs
    assert call_kwargs["verify"] is False


def test_verify_parameter_in_data_request(mock_request):
    """Test that verify parameter is propagated to data requests"""
    with patch("builtins.open", MagicMock()):
        conn = Connector(token="test-token", verify=False)

        # Mock the _validate_response method to avoid errors
        conn._validate_response = MagicMock()

        # Test data request
        conn._data_request("test-id")

        # Check that verify=False was passed to the request
        call_kwargs = mock_request.call_args.kwargs
        assert "verify" in call_kwargs
        assert call_kwargs["verify"] is False


def test_verify_parameter_in_zarr_client(conn):
    """Test that verify parameter is propagated to ZarrClient"""
    # Mock ZarrClient.__init__ to avoid actual initialization
    with patch(
        "oceanum.datamesh.zarr.ZarrClient.__init__", return_value=None
    ) as mock_init:
        # Create a ZarrClient directly with minimal parameters
        ZarrClient(
            connection=conn,
            datasource="test-datasource",
            session=MagicMock(),
            verify=False,
        )

        # Check that verify=False was passed to the ZarrClient constructor
        call_kwargs = mock_init.call_args.kwargs
        assert "verify" in call_kwargs
        assert call_kwargs["verify"] is False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `oceanum` package.

This test suite has been refactored to create test files on the fly rather than
relying on pre-existing fixture files. This approach:
- Eliminates dependencies on external test files
- Ensures test isolation and repeatability
- Provides better cleanup and conflict avoidance
- Makes tests more explicit about their data requirements

Some storage API parameters (limit, match_glob) have known issues that are
handled gracefully with try/except blocks and pytest.skip().
"""
import os
import tempfile
import pytest
import fsspec

from oceanum.storage import FileSystem

REMOTE_PATH = "test_storage"


@pytest.fixture
def fs():
    """Connection fixture"""
    fs = FileSystem(os.environ["DATAMESH_TOKEN"])
    return fs


@pytest.fixture
def fs_async():
    """Connection fixture"""
    fs = FileSystem(os.environ["DATAMESH_TOKEN"], asynchronous=True)
    return fs


def create_test_files():
    """Create temporary test files and return the directory path."""
    tmpdir = tempfile.TemporaryDirectory()

    # Create test files directly in the temp directory
    with open(os.path.join(tmpdir.name, "file1.txt"), "w") as f:
        f.write("hello")
    with open(os.path.join(tmpdir.name, "file2.txt"), "w") as f:
        f.write("world")
    with open(os.path.join(tmpdir.name, "file3.txt"), "w") as f:
        f.write("test content")

    return tmpdir




def test_file_not_found(fs):
    with pytest.raises(FileNotFoundError):
        fs.ls("/not_found")


def test_dir_not_found(fs):
    with pytest.raises(FileNotFoundError):
        fs.ls("/not_found/")


def test_ls(fs):
    """Test basic ls functionality."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_ls'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files directly to the test folder
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_folder}/file2.txt")

        # Test ls functionality
        files = fs.ls(test_folder)
        file_names = [f["name"] for f in files]
        assert f"{test_folder}/file1.txt" in file_names
        assert f"{test_folder}/file2.txt" in file_names

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass

def test_ls_file_prefix(fs):
    """Test ls with file_prefix parameter."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_file_prefix'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files directly to the test folder
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_folder}/file2.txt")
        fs.put(os.path.join(tmpdir.name, "file3.txt"), f"{test_folder}/file3.txt")

        # Test file_prefix filter
        files = fs.ls(test_folder, file_prefix="file1")
        assert len(files) == 1
        assert files[0]["name"] == f"{test_folder}/file1.txt"

        # Test another prefix
        files = fs.ls(test_folder, file_prefix="file")
        assert len(files) == 3  # Should match all files

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass

def test_ls_glob(fs):
    """Test ls with match_glob parameter."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_glob'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files directly to the test folder
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_folder}/file2.txt")
        fs.put(os.path.join(tmpdir.name, "file3.txt"), f"{test_folder}/file3.txt")

        # Verify all files exist first
        files = fs.ls(test_folder)
        assert len(files) == 3

        # Test glob functionality - there's a known issue with the storage API
        # where using match_glob parameter can cause FileNotFoundError
        try:
            files = fs.ls(test_folder, match_glob="**/*2.txt")
            # If this works, verify we get the expected file
            assert len(files) == 1
            assert files[0]["name"] == f"{test_folder}/file2.txt"
        except FileNotFoundError:
            # Known API limitation - match_glob parameter causes issues
            pytest.skip("Storage API has known issue with match_glob parameter")

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass

def test_ls_limit(fs):
    """Test ls with limit parameter."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_limit'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files directly to the test folder
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_folder}/file2.txt")
        fs.put(os.path.join(tmpdir.name, "file3.txt"), f"{test_folder}/file3.txt")

        # Verify all files exist first
        files = fs.ls(test_folder)
        assert len(files) == 3

        # Test limit functionality - there's a known issue with the storage API
        # where using limit parameter can cause FileNotFoundError
        try:
            files = fs.ls(test_folder, limit=2)
            # If this works, verify we get limited results
            assert len(files) <= 2
            assert len(files) >= 1
        except FileNotFoundError:
            # Known API limitation - limit parameter causes issues
            pytest.skip("Storage API has known issue with limit parameter")

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass

def test_get(fs):
    """Test get functionality."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_get'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")

        # Test downloading
        localdir = tempfile.TemporaryDirectory()
        try:
            fs.get(f"{test_folder}/file1.txt", os.path.join(localdir.name, "downloaded_file1.txt"))
            assert os.path.exists(os.path.join(localdir.name, "downloaded_file1.txt"))

            # Verify content
            with open(os.path.join(localdir.name, "downloaded_file1.txt"), "r") as f:
                assert f.read() == "hello"

        finally:
            localdir.cleanup()

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_open(fs):
    """Test fsspec open functionality."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_open'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload file to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")

        # Test opening file via fsspec
        with fsspec.open(f"oceanum://{test_folder}/file1.txt", "r") as f:
            assert f.read() == "hello"

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_copy(fs):
    """Test copy functionality."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_copy'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload file to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")

        # Test copy functionality
        fs.copy(f"{test_folder}/file1.txt", f"{test_folder}/file_copy.txt")

        # Verify copied file exists and has correct content
        with fsspec.open(f"oceanum://{test_folder}/file_copy.txt", "r") as f:
            assert f.read() == "hello"

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_copy_fails(fs):
    """Test copy failure with non-existent file."""
    test_folder = f'{REMOTE_PATH}/test_copy_fail'
    fs.mkdirs(test_folder, exist_ok=True)

    try:
        with pytest.raises(FileNotFoundError):
            fs.copy(
                f"{test_folder}/file_not_there.txt",
                f"{test_folder}/file_copy.txt",
            )
    finally:
        # Clean up remote directory
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_exists_file(fs):
    """Test exists() function with a file that exists."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_exists_file'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload file to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")

        # Test file exists
        assert fs.exists(f"{test_folder}/file1.txt")

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_exists_directory(fs):
    """Test exists() function with a directory that exists."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_exists_dir'
    fs.mkdirs(test_folder, exist_ok=True)

    # Test directory exists
    assert fs.exists(test_folder)

    # Clean up
    try:
        fs.rm(test_folder, recursive=True)
    except:
        pass


def test_exists_nonexistent(fs):
    """Test exists() function with a path that doesn't exist."""
    # Test non-existent file
    assert not fs.exists(f"{REMOTE_PATH}/nonexistent_file_12345.txt")

    # Test non-existent directory
    assert not fs.exists(f"{REMOTE_PATH}/nonexistent_dir_12345")


def test_isfile_true(fs):
    """Test isfile() function with an actual file."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_isfile_true'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_folder}/file2.txt")

        # Test files are correctly identified as files
        assert fs.isfile(f"{test_folder}/file1.txt")
        assert fs.isfile(f"{test_folder}/file2.txt")

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_isfile_false_directory(fs):
    """Test isfile() function with a directory."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_isfile_false'
    fs.mkdirs(test_folder, exist_ok=True)

    try:
        # Test directory is not identified as file
        assert not fs.isfile(test_folder)
        assert not fs.isfile(REMOTE_PATH)

    finally:
        # Clean up remote directory
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_isfile_false_nonexistent(fs):
    """Test isfile() function with non-existent path."""
    # Test non-existent path returns False
    assert not fs.isfile(f"{REMOTE_PATH}/nonexistent_file_12345.txt")


def test_isdir_true(fs):
    """Test isdir() function with an actual directory."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_isdir_true'
    fs.mkdirs(test_folder, exist_ok=True)

    try:
        # Test directory is correctly identified as directory
        assert fs.isdir(test_folder)
        assert fs.isdir(REMOTE_PATH)

    finally:
        # Clean up remote directory
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_isdir_false_file(fs):
    """Test isdir() function with a file."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_isdir_false'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_folder}/file2.txt")

        # Test files are not identified as directories
        assert not fs.isdir(f"{test_folder}/file1.txt")
        assert not fs.isdir(f"{test_folder}/file2.txt")

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_isdir_false_nonexistent(fs):
    """Test isdir() function with non-existent path."""
    # Test non-existent path returns False
    assert not fs.isdir(f"{REMOTE_PATH}/nonexistent_dir_12345")


# Test the sync wrapper functions from oceanum.storage module
def test_sync_exists():
    """Test the sync exists() wrapper function."""
    from oceanum.storage import exists

    # Test with environment token
    token = os.environ.get("DATAMESH_TOKEN")
    if token:
        # Test non-existent path (should be safe to test)
        assert not exists("nonexistent_test_path_12345", token=token)


def test_sync_isfile():
    """Test the sync isfile() wrapper function."""
    from oceanum.storage import isfile

    # Test with environment token
    token = os.environ.get("DATAMESH_TOKEN")
    if token:
        # Test non-existent path (should be safe to test)
        assert not isfile("nonexistent_test_file_12345.txt", token=token)


def test_sync_isdir():
    """Test the sync isdir() wrapper function."""
    from oceanum.storage import isdir

    # Test with environment token
    token = os.environ.get("DATAMESH_TOKEN")
    if token:
        # Test non-existent path (should be safe to test)
        assert not isdir("nonexistent_test_dir_12345", token=token)


def test_exists_edge_cases(fs):
    """Test exists() with various edge cases."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_exists_edge'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload file to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")

        # Test with different path formats
        assert fs.exists(f"{test_folder}/file1.txt")  # Forward slash
        assert fs.exists(os.path.join(test_folder, "file1.txt"))  # os.path.join

        # Test with trailing slash on directory
        assert fs.exists(f"{test_folder}/")
        assert fs.exists(test_folder)  # Without trailing slash

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_file_type_consistency(fs):
    """Test that exists, isfile, and isdir are consistent."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_consistency'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload file to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")

        file_path = f"{test_folder}/file1.txt"
        dir_path = test_folder

        # For existing file
        assert fs.exists(file_path)
        assert fs.isfile(file_path)
        assert not fs.isdir(file_path)

        # For existing directory
        assert fs.exists(dir_path)
        assert not fs.isfile(dir_path)
        assert fs.isdir(dir_path)

        # For non-existent path
        nonexistent = f"{test_folder}/nonexistent_12345"
        assert not fs.exists(nonexistent)
        assert not fs.isfile(nonexistent)
        assert not fs.isdir(nonexistent)

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_rm_file(fs):
    """Test rm() function with a file."""
    # Create unique test directory to avoid conflicts
    test_folder = f'{REMOTE_PATH}/test_rm_file'
    fs.mkdirs(test_folder, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload file to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_folder}/file1.txt")

        file_path = f"{test_folder}/file1.txt"

        # Verify file exists before removal
        assert fs.exists(file_path)

        # Remove the file
        fs.rm(file_path)

        # Verify file no longer exists
        assert not fs.exists(file_path)

    finally:
        tmpdir.cleanup()
        # Clean up remaining remote files
        try:
            fs.rm(test_folder, recursive=True)
        except:
            pass


def test_rm_directory_recursive(fs):
    """Test rm() function with a directory recursively."""
    # Create unique test directory to avoid conflicts
    test_dir = f'{REMOTE_PATH}/test_rm_recursive'
    fs.mkdirs(test_dir, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files to create a directory structure
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_dir}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_dir}/file2.txt")

        # Verify directory and contents exist
        assert fs.exists(test_dir)
        assert fs.isdir(test_dir)
        assert fs.exists(f"{test_dir}/file1.txt")

        # Remove directory recursively
        fs.rm(test_dir, recursive=True)

        # Verify directory no longer exists
        assert not fs.exists(test_dir)

    finally:
        tmpdir.cleanup()


def test_rm_directory_non_recursive_empty(fs):
    """Test rm() function with an empty directory non-recursively."""
    test_dir = os.path.join(REMOTE_PATH, "test_rm_empty")
    fs.mkdirs(test_dir, exist_ok=True)

    # Verify directory exists
    assert fs.exists(test_dir)
    assert fs.isdir(test_dir)

    # Remove empty directory non-recursively
    fs.rm(test_dir, recursive=False)

    # Verify directory no longer exists
    assert not fs.exists(test_dir)


def test_rm_directory_non_recursive_not_empty(fs):
    """Test rm() function with a non-empty directory non-recursively should fail."""
    # Create unique test directory to avoid conflicts
    test_dir = f'{REMOTE_PATH}/test_rm_nonempty'
    fs.mkdirs(test_dir, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files to create a non-empty directory
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_dir}/file1.txt")

        # Verify directory and contents exist
        assert fs.exists(test_dir)
        assert fs.isdir(test_dir)
        assert fs.exists(f"{test_dir}/file1.txt")

        # Attempt to remove non-empty directory non-recursively should raise error
        with pytest.raises(OSError, match="Directory not empty"):
            fs.rm(test_dir, recursive=False)

        # Verify directory still exists
        assert fs.exists(test_dir)

    finally:
        tmpdir.cleanup()
        # Clean up remote files
        try:
            fs.rm(test_dir, recursive=True)
        except:
            pass


def test_rm_nonexistent_path(fs):
    """Test rm() function with non-existent path."""
    nonexistent_path = os.path.join(REMOTE_PATH, "nonexistent_file.txt")

    # Attempt to remove non-existent path should raise FileNotFoundError
    with pytest.raises(FileNotFoundError):
        fs.rm(nonexistent_path)


def test_sync_rm():
    """Test the sync rm() wrapper function."""
    from oceanum.storage import rm

    # Test with environment token
    token = os.environ.get("DATAMESH_TOKEN")
    if token:
        # Test non-existent path (should raise FileNotFoundError)
        with pytest.raises(FileNotFoundError):
            rm("nonexistent_test_file_12345.txt", token=token)


def test_rm_edge_cases(fs):
    """Test rm() with various edge cases."""
    # Create unique test directory to avoid conflicts
    test_dir = f'{REMOTE_PATH}/test_rm_edge'
    fs.mkdirs(test_dir, exist_ok=True)

    # Create test files on the fly
    tmpdir = create_test_files()
    try:
        # Upload files to remote
        fs.put(os.path.join(tmpdir.name, "file1.txt"), f"{test_dir}/file1.txt")
        fs.put(os.path.join(tmpdir.name, "file2.txt"), f"{test_dir}/file2.txt")

        # Remove with forward slash format
        fs.rm(f"{test_dir}/file2.txt")
        assert not fs.exists(f"{test_dir}/file2.txt")

        # Remove remaining file with consistent format
        fs.rm(f"{test_dir}/file1.txt")
        assert not fs.exists(f"{test_dir}/file1.txt")

        # Remove the now-empty directory
        fs.rm(test_dir, recursive=False)
        assert not fs.exists(test_dir)

    finally:
        tmpdir.cleanup()



import datetime
import json
import re
from collections.abc import MutableMapping
import os

import numpy
import xarray
import fsspec
import urllib.parse

from .exceptions import DatameshConnectError, DatameshWriteError
from .session import Session
from .utils import retried_request, DATAMESH_CONNECT_TIMEOUT, DATAMESH_CHUNK_READ_TIMEOUT, DATAMESH_CHUNK_WRITE_TIMEOUT

try:
    import xarray_video as xv

    _VIDEO_SUPPORT = True
except:
    _VIDEO_SUPPORT = False


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def _zarr_proxy(connection, datasource_id, parameters={}):
    try:
        mapper = fsspec.get_mapper(
            f"{connection._gateway}/zarr/{datasource_id}",
            headers={
                **connection._auth_headers,
                "X-PARAMETERS": json.dumps(parameters, default=json_serial),
            },
        )
    except Exception as e:
        raise DatameshConnectError(str(e))
    return mapper


class ZarrClient(MutableMapping):
    def __init__(
        self,
        connection,
        datasource,
        session,
        parameters={},
        method="post",
        retries=10,
        read_timeout=DATAMESH_CHUNK_READ_TIMEOUT,
        connect_timeout=DATAMESH_CONNECT_TIMEOUT,
        write_timeout=DATAMESH_CHUNK_WRITE_TIMEOUT,
        nocache=False,
        api="query",
        reference_id=None,
        verify=True,
        storage_backend=None
    ):
        self.datasource = datasource
        self.session = session
        self.method = method
        self._is_v1 = connection._is_v1
        self.api = api if connection._is_v1 else "zarr"
        self.headers = {**connection._auth_headers}
        self.headers = session.add_header(self.headers)
        if nocache:
            self.headers["cache-control"] = "no-transform,no-cache"
        if parameters:
            self.headers["X-PARAMETERS"] = json.dumps(parameters)
        if self.api == "zarr":
            self._proxy = connection._gateway + "/zarr"
        elif self.api == "query":
            self._proxy = connection._gateway + "/zarr/query"
        else:
            raise DatameshConnectError(f"Unknown api: {self.api}")
        self.retries = retries
        self.read_timeout = read_timeout
        self.connect_timeout = connect_timeout
        self.write_timeout = write_timeout
        self.verify = verify
        if storage_backend is not None:
            self.headers["X-DATAMESH-STORAGE-BACKEND"] = storage_backend

    def _retried_request(
        self,
        path,
        method="GET",
        data=None,
        connect_timeout=DATAMESH_CONNECT_TIMEOUT,
        read_timeout=DATAMESH_CHUNK_READ_TIMEOUT,
    ):
        resp = retried_request(
            url=path,
            method=method,
            data=data,
            headers=self.headers,
            retries=self.retries,
            timeout=(connect_timeout, read_timeout),
            verify=self.verify,
        )
        if resp.status_code == 401:
            raise DatameshConnectError(f"Not Authorized {resp.text}")
        if resp.status_code == 410:
            raise DatameshConnectError(f"Datasource no longer exists or was deleted within your session")
        if resp.status_code >= 500:
            raise DatameshConnectError(
                f"Server error {resp.status_code}: {resp.text}"
            )
        return resp

    def __getitem__(self, item):
        encoded_item = urllib.parse.quote(item, safe='/')
        resp = self._retried_request(
            f"{self._proxy}/{self.datasource}/{encoded_item}",
            connect_timeout=self.connect_timeout,
            read_timeout=self.read_timeout,
        )
        if resp.status_code >= 300:
            raise KeyError(item)
        return resp.content

    def __contains__(self, item):
        encoded_item = urllib.parse.quote(item, safe='/')
        resp = self._retried_request(
            f"{self._proxy}/{self.datasource}/{encoded_item}",
            method="HEAD" if self._is_v1 else "GET",
            connect_timeout=self.connect_timeout,
            read_timeout=self.read_timeout,
        )
        if resp.status_code != 200:
            return False
        return True

    def __setitem__(self, item, value):
        if self.api == "query":
            raise DatameshConnectError("Query api does not support write operations")
        encoded_item = urllib.parse.quote(item, safe='/')
        res = self._retried_request(
            f"{self._proxy}/{self.datasource}/{encoded_item}",
            method=self.method,
            data=value,
            connect_timeout=self.write_timeout,
            read_timeout=self.write_timeout,
        )
        if res.status_code >= 300:
            raise DatameshWriteError(
                f"Failed to write {item}: {res.status_code} - {res.text}"
            )

    def __delitem__(self, item):
        if self.api == "query":
            raise DatameshConnectError("Query api does not support delete operations")
        encoded_item = urllib.parse.quote(item, safe='/')
        self._retried_request(
            f"{self._proxy}/{self.datasource}/{encoded_item}",
            method="DELETE",
            connect_timeout=self.connect_timeout,
            read_timeout=10,
        )

    def __iter__(self):
        resp = self._retried_request(
            f"{self._proxy}/{self.datasource}/",
            connect_timeout=self.connect_timeout,
            read_timeout=self.read_timeout,
        )
        if not resp:
            return
        ex = re.compile(r"""<(a|A)\s+(?:[^>]*?\s+)?(href|HREF)=["'](?P<url>[^"']+)""")
        links = [u[2] for u in ex.findall(resp.text)]
        for link in links:
            yield link

    def __len__(self):
        return 0

    def clear(self):
        self.__delitem__("")


def _to_zarr(data, store, **kwargs):
    if _VIDEO_SUPPORT:
        data.video.to_zarr(store, **kwargs)
    else:
        data.to_zarr(store, **kwargs)


def zarr_write(connection, datasource_id, data, append=None, overwrite=False):
    with Session.acquire(connection) as session:
        store = ZarrClient(connection, datasource_id, session, api="zarr", nocache=True)
        if overwrite is True:
            store.clear()
            append = None
        else:
            ds = connection.get_datasource(datasource_id)
        if append and ds._exists:
            if append not in ds.dataschema.coords:
                raise DatameshWriteError(
                    f"Append coordinate {append} not in existing zarr"
                )
            with xarray.open_zarr(store) as dexist:
                cexist = dexist[append]
                if len(cexist.dims) > 1:
                    raise DatameshWriteError(
                        f"Append coordinate {append} has more than one dimension"
                    )
                append_dim = cexist.dims[0]
                (replace_range,) = numpy.nonzero(
                    ((cexist >= data[append][0]) & (cexist <= data[append][-1])).values
                )  # Get range in new data which overlaps - this just replaces everything >= first value in the new data
                if len(replace_range):
                    # Fail if the replacement range is larger than incomign data
                    if len(replace_range) > len(data[append]):
                        raise DatameshWriteError(
                            f"Cannot append to zarr with a region that would be smaller than the original"
                        )

                    drop_coords = [c for c in data.coords if c != append]
                    drop_vars = [
                        v for v in data.data_vars if not append in data[v].dims
                    ]
                    replace_section = data.isel(
                        **{append_dim: slice(0, len(replace_range))}
                    ).drop(drop_coords + drop_vars)
                    replace_slice = slice(replace_range[0], replace_range[-1] + 1)
                    # Fail if we are replacing an internal section and ends of coordinates do not match
                    if replace_range[-1] + 1 < len(cexist) and not numpy.array_equal(
                        replace_section[append], cexist[replace_slice]
                    ):
                        raise DatameshWriteError(
                            f"Data inconsistency on coordinate {append} replacing a inner section of an existing zarr array"
                        )
                    _to_zarr(
                        replace_section,
                        store,
                        mode="a",
                        region={append_dim: replace_slice},
                    )
                if len(data[append]) > len(replace_range):
                    append_chunk = data.isel(
                        **{append_dim: slice(len(replace_range), None)}
                    )
                    _to_zarr(
                        append_chunk,
                        store,
                        mode="a",
                        append_dim=append_dim,
                        consolidated=True,
                    )
        else:
            _to_zarr(data, store, mode="w", consolidated=True)
            ds = connection.get_datasource(datasource_id)
            ds.dataschema = data.to_dict(data=False)
        return ds



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `oceanum storage` CLI commands.

These tests require authentication with the Oceanum service. They will be automatically
skipped in CI/CD environments or when the user is not logged in locally.

To run these tests locally:
1. Ensure you have the DATAMESH_TOKEN environment variable set
2. Run `oceanum auth login` to authenticate
3. Run the tests with `pytest tests/test_cli_storage.py`

In CI/CD environments, these tests will be skipped and only basic help command tests
will run to ensure the CLI is properly installed and accessible.
"""
import os
import sys
import tempfile
import pytest
from unittest.mock import patch, MagicMock
from functools import update_wrapper

from click.testing import CliRunner
from oceanum.cli import main as cli_main, storage as storage_cli
from oceanum.cli.models import TokenResponse
from oceanum.cli.storage import *
from oceanum.storage import FileSystem

REMOTE_PATH = "test_storage_cli"

@pytest.fixture
def fs():
    return FileSystem(os.environ["DATAMESH_TOKEN"])


@pytest.fixture
def test_files(fs):
    """Create test files and directories for CLI testing."""
    tmpdir = tempfile.TemporaryDirectory()

    # Create test file structure
    os.makedirs(os.path.join(tmpdir.name, "test_dir"))
    with open(os.path.join(tmpdir.name, "test_file.txt"), "w") as f:
        f.write("test content")
    with open(os.path.join(tmpdir.name, "test_dir", "nested_file.txt"), "w") as f:
        f.write("nested content")

    # Upload to remote storage
    fs.mkdirs(REMOTE_PATH, exist_ok=True)
    fs.put(os.path.join(tmpdir.name, "test_file.txt"), f"{REMOTE_PATH}/test_file.txt")
    fs.put(os.path.join(tmpdir.name, "test_dir"), REMOTE_PATH, recursive=True)

    yield {
        "file": f"{REMOTE_PATH}/test_file.txt",
        "dir": f"{REMOTE_PATH}/test_dir",
        "nested_file": f"{REMOTE_PATH}/test_dir/nested_file.txt",
        "nonexistent": f"{REMOTE_PATH}/nonexistent_path",
    }

    tmpdir.cleanup()


def run_cli_command(args):
    """Helper function to run CLI commands."""
    runner = CliRunner()
    test_token = TokenResponse(
        access_token='test_access_token',
        refresh_token='test_refresh_token',
        expires_in=3600,
        token_type='Bearer',
    )
    # Mock the token so we by-pass login_required decorator
    with patch.object(TokenResponse, 'load', return_value=test_token):
        # Mock the get_bearer_token function to return None so will use DATAMESH_TOKEN
        with patch.object(storage_cli, 'get_bearer_token', return_value=None):
            result = runner.invoke(cli_main, args)
            if result.exception:
                print(f"Error running command '{' '.join(args)}': {result.exception}", file=sys.stderr)
    return result


def test_cli_exists_file(test_files):
    """Test CLI exists command with existing file."""
    result = run_cli_command(["storage", "exists", test_files["file"]])
    assert result.exit_code == 0
    assert "EXISTS:" in result.stdout
    assert test_files["file"] in result.stdout


def test_cli_exists_directory(test_files):
    """Test CLI exists command with existing directory."""
    result = run_cli_command(["storage", "exists", test_files["dir"]])
    assert result.exit_code == 0
    assert "EXISTS:" in result.stdout
    assert test_files["dir"] in result.stdout


def test_cli_exists_nonexistent(test_files):
    """Test CLI exists command with non-existent path."""
    result = run_cli_command(["storage", "exists", test_files["nonexistent"]])
    assert result.exit_code == 1
    assert "NOT FOUND:" in result.stdout
    assert test_files["nonexistent"] in result.stdout


def test_cli_isfile_true(test_files):
    """Test CLI isfile command with actual file."""
    result = run_cli_command(["storage", "isfile", test_files["file"]])
    assert result.exit_code == 0
    assert "FILE:" in result.stdout
    assert test_files["file"] in result.stdout


def test_cli_isfile_false_directory(test_files):
    """Test CLI isfile command with directory."""
    result = run_cli_command(["storage", "isfile", test_files["dir"]])
    assert result.exit_code == 1
    assert "NOT A FILE:" in result.stdout
    assert test_files["dir"] in result.stdout


def test_cli_isdir_true(test_files):
    """Test CLI isdir command with actual directory."""
    result = run_cli_command(["storage", "isdir", test_files["dir"]])
    assert result.exit_code == 0
    assert "DIRECTORY:" in result.stdout
    assert test_files["dir"] in result.stdout


def test_cli_isdir_false_file(test_files):
    """Test CLI isdir command with file."""
    result = run_cli_command(["storage", "isdir", test_files["file"]])
    assert result.exit_code == 1
    assert "NOT A DIRECTORY:" in result.stdout
    assert test_files["file"] in result.stdout


def test_cli_rm_file_force(fs):
    """Test CLI rm command with force flag (no confirmation)."""
    # Create a test file to remove
    test_path = f"{REMOTE_PATH}/rm_test_file.txt"
    tmpdir = tempfile.TemporaryDirectory()
    with open(os.path.join(tmpdir.name, "rm_test.txt"), "w") as f:
        f.write("test")
    fs.put(os.path.join(tmpdir.name, "rm_test.txt"), test_path)

    # Verify file exists
    assert fs.exists(test_path)
    # Remove with force flag
    result = run_cli_command(["storage", "rm", "-f", test_path])
    assert result.exit_code == 0
    assert "Successfully removed:" in result.stdout
    assert test_path in result.stdout

    # Verify file is gone
    assert not fs.exists(test_path)

    tmpdir.cleanup()


def test_cli_ls_command():
    """Test CLI ls command with default parameters."""
    result = run_cli_command(["storage", "ls", REMOTE_PATH])
    assert result.exit_code == 0
    assert "test_file.txt" in result.stdout
    assert "test_dir" in result.stdout

def test_cli_help_commands():
    """Test that help text is available for all commands."""
    # Test main storage help
    result = run_cli_command(["storage", "--help"])
    assert result.exit_code == 0
    assert "exists" in result.stdout
    assert "isfile" in result.stdout
    assert "isdir" in result.stdout
    assert "rm" in result.stdout

    # Test individual command help
    for cmd in ["exists", "isfile", "isdir", "rm"]:
        result = run_cli_command(["storage", cmd, "--help"])
        assert result.exit_code == 0
        assert "Usage:" in result.stdout


def test_cli_rm_help_shows_new_options():
    """Test that rm help shows the new force option."""
    result = run_cli_command(["storage", "rm", "--help"])
    assert result.exit_code == 0
    assert "-f, --force" in result.stdout
    assert "Force removal without confirmation" in result.stdout
    assert "-r, --recursive" in result.stdout


def test_cli_storage_help():
    """Test that storage help works without login."""
    result = run_cli_command(["storage", "--help"])
    assert result.exit_code == 0
    assert "Usage:" in result.stdout
    # Only test for commands that actually exist
    assert "ls" in result.stdout
    assert "get" in result.stdout
    assert "put" in result.stdout
    assert "rm" in result.stdout



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for LocalCache class."""
import os
import time
import tempfile
import pytest
import pandas as pd
import geopandas as gpd
import xarray as xr
import numpy as np
from shapely.geometry import Point

from oceanum.datamesh.cache import LocalCache, CacheError
from oceanum.datamesh.query import Query


@pytest.fixture
def temp_cache_dir():
    """Create a temporary cache directory for testing."""
    temp_dir = tempfile.mkdtemp(prefix="test_cache_")
    yield temp_dir
    # Cleanup after tests
    import shutil

    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)


@pytest.fixture
def cache(temp_cache_dir):
    """Create a LocalCache instance with temporary directory."""
    return LocalCache(cache_timeout=2, cache_dir=temp_cache_dir, lock_timeout=2)


@pytest.fixture
def sample_query():
    """Create a sample query for testing."""
    return Query(datasource="test-datasource")


@pytest.fixture
def sample_xarray_dataset():
    """Create a sample xarray Dataset."""
    return xr.Dataset(
        {
            "temperature": (["x", "y"], np.random.rand(3, 3)),
            "pressure": (["x", "y"], np.random.rand(3, 3)),
        },
        coords={"x": [0, 1, 2], "y": [0, 1, 2]},
    )


@pytest.fixture
def sample_geodataframe():
    """Create a sample GeoDataFrame."""
    return gpd.GeoDataFrame(
        {"name": ["A", "B", "C"], "value": [1, 2, 3]},
        geometry=[Point(0, 0), Point(1, 1), Point(2, 2)],
        crs="EPSG:4326",
    )


@pytest.fixture
def sample_dataframe():
    """Create a sample DataFrame."""
    return pd.DataFrame({"col1": [1, 2, 3], "col2": ["a", "b", "c"]})


class TestLocalCacheInit:
    """Test LocalCache initialization."""

    def test_init_creates_cache_dir(self, temp_cache_dir):
        """Test that cache directory is created if it doesn't exist."""
        cache_dir = os.path.join(temp_cache_dir, "new_cache")
        assert not os.path.exists(cache_dir)
        cache = LocalCache(cache_dir=cache_dir)
        assert os.path.exists(cache_dir)
        assert cache.cache_dir == cache_dir

    def test_init_with_existing_dir(self, temp_cache_dir):
        """Test initialization with existing directory."""
        cache = LocalCache(cache_dir=temp_cache_dir)
        assert cache.cache_dir == temp_cache_dir

    def test_init_default_values(self, temp_cache_dir):
        """Test default initialization values."""
        cache = LocalCache(cache_dir=temp_cache_dir)
        assert cache.cache_timeout == 600
        assert cache.lock_timeout == 60

    def test_init_custom_values(self, temp_cache_dir):
        """Test custom initialization values."""
        cache = LocalCache(cache_timeout=300, cache_dir=temp_cache_dir, lock_timeout=30)
        assert cache.cache_timeout == 300
        assert cache.lock_timeout == 30


class TestCachePath:
    """Test cache path generation."""

    def test_cachepath_with_query_object(self, cache, sample_query):
        """Test cache path generation with Query object."""
        path = cache._cachepath(sample_query)
        assert path.startswith(cache.cache_dir)
        assert len(os.path.basename(path)) == 56  # SHA224 hex digest length

    def test_cachepath_with_dict(self, cache):
        """Test cache path generation with dictionary."""
        query_dict = {"datasource": "test-datasource"}
        path = cache._cachepath(query_dict)
        assert path.startswith(cache.cache_dir)

    def test_cachepath_consistency(self, cache, sample_query):
        """Test that same query produces same cache path."""
        path1 = cache._cachepath(sample_query)
        path2 = cache._cachepath(sample_query)
        assert path1 == path2

    def test_cachepath_different_queries(self, cache):
        """Test that different queries produce different cache paths."""
        query1 = Query(datasource="datasource1")
        query2 = Query(datasource="datasource2")
        path1 = cache._cachepath(query1)
        path2 = cache._cachepath(query2)
        assert path1 != path2


class TestLocking:
    """Test cache locking mechanism."""

    def test_lock_creates_lockfile(self, cache, sample_query):
        """Test that lock creates a lockfile."""
        lockfile = cache._cachepath(sample_query) + ".lock"
        assert not os.path.exists(lockfile)
        cache.lock(sample_query)
        assert os.path.exists(lockfile)

    def test_locked_returns_true_when_locked(self, cache, sample_query):
        """Test that _locked returns True when file is locked."""
        assert not cache._locked(sample_query)
        cache.lock(sample_query)
        assert cache._locked(sample_query)

    def test_locked_returns_false_after_timeout(self, cache, sample_query):
        """Test that lock expires after timeout."""
        cache.lock(sample_query)
        assert cache._locked(sample_query)
        time.sleep(2.1)  # Wait for lock timeout
        assert not cache._locked(sample_query)

    def test_unlock_removes_lockfile(self, cache, sample_query):
        """Test that unlock removes the lockfile."""
        cache.lock(sample_query)
        lockfile = cache._cachepath(sample_query) + ".lock"
        assert os.path.exists(lockfile)
        cache.unlock(sample_query)
        assert not os.path.exists(lockfile)

    def test_unlock_when_not_locked(self, cache, sample_query):
        """Test that unlock does nothing when not locked."""
        cache.unlock(sample_query)  # Should not raise error

    def test_lock_idempotent(self, cache, sample_query):
        """Test that locking multiple times is safe."""
        cache.lock(sample_query)
        cache.lock(sample_query)  # Should not raise error
        assert cache._locked(sample_query)


class TestPutAndGet:
    """Test cache put and get operations."""

    def test_put_and_get_xarray_dataset(
        self, cache, sample_query, sample_xarray_dataset
    ):
        """Test storing and retrieving xarray Dataset."""
        cache.put(sample_query, sample_xarray_dataset)
        cached_file = cache._cachepath(sample_query) + ".nc"
        assert os.path.exists(cached_file)

        retrieved = cache.get(sample_query)
        assert isinstance(retrieved, xr.Dataset)
        xr.testing.assert_equal(retrieved, sample_xarray_dataset)

    def test_put_and_get_geodataframe(self, cache, sample_query, sample_geodataframe):
        """Test storing and retrieving GeoDataFrame."""
        cache.put(sample_query, sample_geodataframe)
        cached_file = cache._cachepath(sample_query) + ".gpq"
        assert os.path.exists(cached_file)

        retrieved = cache.get(sample_query)
        assert isinstance(retrieved, gpd.GeoDataFrame)
        pd.testing.assert_frame_equal(retrieved, sample_geodataframe)

    def test_put_and_get_dataframe(self, cache, sample_query, sample_dataframe):
        """Test storing and retrieving DataFrame."""
        cache.put(sample_query, sample_dataframe)
        cached_file = cache._cachepath(sample_query) + ".pq"
        assert os.path.exists(cached_file)

        retrieved = cache.get(sample_query)
        assert isinstance(retrieved, pd.DataFrame)
        pd.testing.assert_frame_equal(retrieved, sample_dataframe)

    def test_put_unsupported_type(self, cache, sample_query):
        """Test that putting unsupported type raises TypeError."""
        with pytest.raises(TypeError, match="Unsupported data type"):
            cache.put(sample_query, "unsupported string data")

    def test_get_nonexistent_cache(self, cache, sample_query):
        """Test getting from cache when no cached data exists."""
        result = cache.get(sample_query)
        assert result is None

    def test_get_expired_cache(self, cache, sample_query, sample_dataframe):
        """Test that expired cache returns None and removes file."""
        cache.put(sample_query, sample_dataframe)
        cached_file = cache._cachepath(sample_query) + ".pq"
        assert os.path.exists(cached_file)

        time.sleep(2.1)  # Wait for cache timeout
        result = cache.get(sample_query)
        assert result is None
        assert not os.path.exists(cached_file)

    def test_get_with_lock_waits(self, cache, sample_query, sample_dataframe):
        """Test that get waits when cache is locked."""
        cache.lock(sample_query)

        # Start time
        start = time.time()

        # In a real scenario, another process would unlock
        # For testing, we'll unlock after a short delay
        import threading

        def unlock_after_delay():
            time.sleep(0.3)
            cache.unlock(sample_query)

        thread = threading.Thread(target=unlock_after_delay)
        thread.start()

        result = cache.get(sample_query, timeout=5)
        elapsed = time.time() - start

        thread.join()

        # Should have waited at least 0.3 seconds
        assert elapsed >= 0.3
        assert result is None  # No cached data exists

    def test_get_with_lock_timeout(self, cache, sample_query):
        """Test that get raises CacheError when lock timeout is reached."""
        cache.lock(sample_query)

        assert cache.get(sample_query, timeout=0.5) is None

    def test_get_corrupted_cache_returns_none(self, cache, sample_query):
        """Test that corrupted cache file returns None."""
        cached_file = cache._cachepath(sample_query) + ".nc"

        # Create a corrupted file
        with open(cached_file, "w") as f:
            f.write("corrupted data")

        result = cache.get(sample_query)
        assert result is None


class TestCopy:
    """Test cache copy operation."""

    def test_copy_moves_file(self, cache, sample_query, temp_cache_dir):
        """Test that copy moves file to cache location."""
        source_file = os.path.join(temp_cache_dir, "source.nc")

        # Create a source file
        with open(source_file, "w") as f:
            f.write("test data")

        assert os.path.exists(source_file)

        cache.copy(sample_query, source_file, ".nc")

        cached_file = cache._cachepath(sample_query) + ".nc"
        assert os.path.exists(cached_file)
        assert not os.path.exists(source_file)  # Original should be moved


class TestCacheIntegration:
    """Integration tests for cache operations."""

    def test_multiple_queries_different_caches(self, cache, sample_dataframe):
        """Test that different queries create different cache files."""
        query1 = Query(datasource="datasource1")
        query2 = Query(datasource="datasource2")

        cache.put(query1, sample_dataframe)
        cache.put(query2, sample_dataframe)

        cached_file1 = cache._cachepath(query1) + ".pq"
        cached_file2 = cache._cachepath(query2) + ".pq"

        assert os.path.exists(cached_file1)
        assert os.path.exists(cached_file2)
        assert cached_file1 != cached_file2

    def test_cache_overwrite(self, cache, sample_query):
        """Test that putting new data overwrites old cache."""
        df1 = pd.DataFrame({"col": [1, 2, 3]})
        df2 = pd.DataFrame({"col": [4, 5, 6]})

        cache.put(sample_query, df1)
        retrieved1 = cache.get(sample_query)
        pd.testing.assert_frame_equal(retrieved1, df1)

        cache.put(sample_query, df2)
        retrieved2 = cache.get(sample_query)
        pd.testing.assert_frame_equal(retrieved2, df2)

    def test_lock_unlock_workflow(self, cache, sample_query, sample_dataframe):
        """Test typical lock-unlock workflow."""
        # Lock before writing
        cache.lock(sample_query)
        assert cache._locked(sample_query)

        # Write data
        cache.put(sample_query, sample_dataframe)

        # Unlock after writing
        cache.unlock(sample_query)
        assert not cache._locked(sample_query)

        # Should be able to read
        result = cache.get(sample_query)
        assert result is not None
        pd.testing.assert_frame_equal(result, sample_dataframe)



from unittest import TestCase
import json
import yaml

from oceanum.cli.renderer import Renderer, RenderField

class RendererTests(TestCase):
    
    def test_simple_dict_data(self):
        data = {
            'name' : 'SlimShady',
            'status': 'hidden',
            'notshown': 'notshown'
        }
        fields = [
            RenderField(label='Name', path='$.name'),
        ]
        renderer = Renderer(data, fields, output='table')
        table_output =  renderer.render_table()
        assert 'Name' in table_output
        assert 'SlimShady' in table_output
        assert 'notshown' not in table_output
        json_output = renderer.render_json()
        assert json.loads(json_output.strip("'")) == [data]
        yaml_output = renderer.render_yaml()
        assert yaml.safe_load(yaml_output) == [data]


    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `oceanum` package."""
import os
import pytest
import pandas
import geopandas
import pyproj
import xarray
import rioxarray
import numpy
import dask.dataframe

from click.testing import CliRunner

from oceanum.datamesh import Connector, Datasource
from oceanum.datamesh.exceptions import DatameshWriteError
from oceanum import cli

HERE = os.path.dirname(__file__)


@pytest.fixture
def conn():
    """Connection fixture"""
    return Connector()


@pytest.fixture
def dataframe():
    df = pandas.read_csv(
        os.path.join(HERE, "data", "point_data_1.csv"),
        parse_dates=True,
        index_col=0,
    )
    return df


@pytest.fixture
def dataset():
    ds = xarray.open_dataset(os.path.join(HERE, "data", "grid_data_1.nc"))
    return ds


@pytest.fixture
def geotiff():
    ds = xarray.open_dataset(os.path.join(HERE, "data", "raster_data_1.tif"))
    return ds


@pytest.fixture
def mark_for_cleanup():
    """Fixture to track and cleanup datasources created during tests."""
    created_datasources = []
    
    def register_datasource(datasource_id, conn):
        created_datasources.append((datasource_id, conn))
        return datasource_id
    
    yield register_datasource
    
    # Cleanup all registered datasources
    for datasource_id, conn in created_datasources:
        try:
            conn.delete_datasource(datasource_id)
        except Exception as e:
            pass


def test_write_dataframe(conn, dataframe, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataframe", conn)
    conn.write_datasource(
        datasource_id,
        dataframe,
        {"type": "Point", "coordinates": [174, -39]},
        overwrite=True,
    )
    df = conn.load_datasource(datasource_id)
    assert (df == dataframe).all().all()
    conn.delete_datasource(datasource_id)


def test_write_dask_dataframe(conn, dataframe, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dask-dataframe", conn)
    conn.write_datasource(
        datasource_id,
        dask.dataframe.from_pandas(dataframe, npartitions=1),
        {"type": "Point", "coordinates": [174, -39]},
        overwrite=True,
    )
    df = conn.load_datasource(datasource_id, use_dask=True)
    assert isinstance(df, xarray.Dataset)
    df = df.compute()
    assert (df["u10"] == dataframe["u10"]).all()
    conn.delete_datasource(datasource_id)


def test_write_dataset(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset", conn)
    conn.write_datasource(datasource_id, dataset, overwrite=True)
    ds = conn.load_datasource(datasource_id)
    assert (ds == dataset).all()["u10"]
    conn.delete_datasource(datasource_id)


def test_write_dataset_guess(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset-guess", conn)
    conn.write_datasource(
        datasource_id,
        dataset,
        overwrite=True,
        coordinates={"t": "time", "x": "longitude", "y": "latitude"},
    )
    dsrc = conn.get_datasource(datasource_id)
    assert dsrc.geom.bounds == (173, -38, 174, -37)
    conn.delete_datasource(datasource_id)


def test_write_dataset_crs(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset-crs", conn)
    dataset_2193 = dataset.copy().rename(
        {"longitude": "easting", "latitude": "northing"}
    )
    x, y = pyproj.Transformer.from_crs(
        pyproj.CRS("EPSG:4326"), pyproj.CRS("EPSG:2193"), always_xy=True
    ).transform(dataset["longitude"], dataset["latitude"])
    dataset_2193["easting"] = x
    dataset_2193["northing"] = y
    conn.write_datasource(
        datasource_id,
        dataset_2193,
        overwrite=True,
        coordinates={"t": "time", "x": "easting", "y": "northing"},
        crs=2193,
    )
    dsrc = conn.get_datasource(datasource_id)
    assert dsrc.geom.bounds[0] == 173
    conn.delete_datasource(datasource_id)


def test_bad_coordinates_fail(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset-coord-fail", conn)
    with pytest.raises(DatameshWriteError):
        conn.write_datasource(
            datasource_id,
            dataset,
            overwrite=True,
            coordinates={
                "t": "time",
                "x": "longitude",
                "y": "latitude",
                "z": "i_do_not_exist",
            },
        )
    conn.delete_datasource(datasource_id)


def test_append_dataset(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset-append", conn)
    dataset2 = dataset.copy()
    dataset2["time"] = dataset["time"] + numpy.timedelta64(1, "D")
    conn.write_datasource(datasource_id, dataset, overwrite=True)
    conn.write_datasource(datasource_id, dataset2, append="time")
    ds = conn.load_datasource(datasource_id)
    assert len(ds["u10"]) == 73
    assert (ds["u10"][:10] == dataset["u10"][:10]).all()
    conn.delete_datasource(datasource_id)


def test_write_region_dataset(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset-region", conn)
    dataset2 = dataset.isel(time=slice(10,-10))+2
    conn.write_datasource(datasource_id, dataset, overwrite=True)
    conn.write_datasource(datasource_id, dataset2, append="time")
    ds = conn.load_datasource(datasource_id)
    assert len(ds["u10"]) == 49
    assert (ds["u10"][:10] == dataset["u10"][:10]).all()
    numpy.testing.assert_almost_equal(ds["u10"][10:-10].values,
                                      (dataset["u10"][10:-10]+2).values,
                                      decimal=3)
    assert (ds["u10"][-10:] == dataset["u10"][-10:]).all()
    conn.delete_datasource(datasource_id)


def test_write_region_chunked_dataset(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset-region-chunked", conn)
    dataset2 = dataset.isel(time=slice(10,-10))+2
    conn.write_datasource(datasource_id, dataset.chunk({"time": 7}), overwrite=True)
    conn.write_datasource(datasource_id, dataset2, append="time")
    ds = conn.load_datasource(datasource_id)
    assert len(ds["u10"]) == 49
    assert (ds["u10"][:10] == dataset["u10"][:10]).all()
    numpy.testing.assert_almost_equal(ds["u10"][10:-10].values,
                                      (dataset["u10"][10:-10]+2).values,
                                      decimal=3)
    assert (ds["u10"][-10:] == dataset["u10"][-10:]).all()
    conn.delete_datasource(datasource_id)


def test_append_dataset_fail(conn, dataset, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-dataset-fail", conn)
    dataset2 = dataset.copy()
    dataset2["time"] = pandas.date_range(
        dataset["time"][10].values, dataset["time"][20].values, 49
    ).values
    conn.write_datasource(datasource_id, dataset, overwrite=True)
    with pytest.raises(DatameshWriteError):
        conn.write_datasource(datasource_id, dataset2, append="time")
    conn.delete_datasource(datasource_id)


def test_write_metadata(conn, dataframe, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-metadata", conn)
    conn.write_datasource(
        datasource_id,
        None,
        name=datasource_id,
        coordinates={},
        driver="null",
        geometry={"type": "Point", "coordinates": [174, -39]},
        schema={"attrs": {}, "dims": {}, "coords": {}, "data_vars": {}},
        tstart="2020-01-01T00:00:00Z",
    )
    ds = conn.get_datasource(datasource_id)
    assert ds.name == datasource_id
    conn.delete_datasource(datasource_id)


def test_update_metadata(conn, dataframe, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-update-metadata", conn)
    conn.write_datasource(
        datasource_id, dataframe, {"type": "Point", "coordinates": [174, -39]}
    )
    conn.write_datasource(
        datasource_id,
        None,
        name="new name",
    )
    ds = conn.get_datasource(datasource_id)
    assert ds.name == "new name"
    conn.delete_datasource(datasource_id)


def test_write_metadata_with_crs(conn, dataframe, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-metadata-crs", conn)
    conn.write_datasource(
        datasource_id,
        None,
        name=datasource_id,
        coordinates={},
        driver="null",
        geometry={"type": "Point", "coordinates": [1686592, 5682747]},
        tstart="2020-01-01T00:00:00Z",
        crs=2193,
    )
    ds = conn.get_datasource(datasource_id)
    assert ds.dataschema.attrs["crs"] == 2193
    assert abs(ds.geom.x - 174) < 1e-4
    conn.delete_datasource(datasource_id)


def test_write_metadata_with_label(conn, dataframe, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-metadata-label", conn)
    conn.write_datasource(
        datasource_id,
        None,
        name=datasource_id,
        coordinates={},
        driver="null",
        geometry={"type": "Point", "coordinates": [10, -15]},
        tstart="2020-01-01T00:00:00Z",
        labels=["test_label"],
    )
    ds = conn.get_datasource(datasource_id)
    assert ds.labels[0] == "test_label"
    assert len(ds.labels) == 1
    conn.delete_datasource(datasource_id)


def test_write_metadata_with_bad_crs(conn, dataframe, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-metadata-bad-crs", conn)
    with pytest.raises(DatameshWriteError):
        conn.write_datasource(
            datasource_id,
            None,
            name=datasource_id,
            coordinates={},
            driver="null",
            geometry={"type": "Point", "coordinates": [1686592, 5682747]},
            tstart="2020-01-01T00:00:00Z",
        )


def test_write_raster(conn, geotiff, mark_for_cleanup):
    datasource_id = mark_for_cleanup("test-write-raster", conn)
    conn.write_datasource(datasource_id, geotiff, overwrite=True)
    ds = conn.load_datasource(datasource_id)
    assert ds["band"]
    conn.delete_datasource(datasource_id)



from pydantic import BaseModel
from typing import Optional
from datetime import datetime, timedelta
from .exceptions import DatameshConnectError, DatameshSessionError
from .utils import retried_request
import atexit
import os


class Session(BaseModel):
    id: str
    user: str
    creation_time: datetime
    end_time: datetime
    write: bool
    allow_multiwrite: bool = False
    verified: bool = False

    @classmethod
    def acquire(
        cls,
        connection,
        allow_multiwrite: Optional[bool] = False,
        duration: Optional[int] = None,
    ):
        """
        Acquire a session from the connection.

        Parameters
        ----------
        connection : Connection
            Connection object to acquire session from.
        allow_multiwrite : bool
            Whether to allow other sessions to write to datasource
            already being written to by this session.
            Default is False.
        duration: int
            The duration of the session in seconds. Will default to
            the connection session duration if set or 3600 (1 hour)
        """

        # Back-compatibility with beta version (returning dummy session object)
        if not connection._is_v1:
            session = cls(
                id="dummy_session",
                user="dummy_user",
                creation_time=datetime.now(),
                end_time=datetime.now()
                + timedelta(seconds=connection._session_params.get("duration", 3600)),
                write=False,
                verified=False,
            )
            session._connection = connection
            atexit.register(session.close)
            return session
        # v1
        try:
            headers = connection._auth_headers.copy()
            headers["Cache-Control"] = "no-store"
            params = connection._session_params.copy()
            params["allow_multiwrite"] = allow_multiwrite
            if duration is not None:
                params["duration"] = duration
            res = retried_request(
                f"{connection._gateway}/session/", params=params, headers=headers
            )
            if res.status_code != 200:
                raise DatameshConnectError(
                    "Failed to create session with error: " + res.text
                )
            session = cls(**res.json())
            session._connection = connection
            atexit.register(session.close)
            return session
        except Exception as e:
            raise DatameshSessionError(f"Error when acquiring datamesh session {e}")

    @classmethod
    def from_proxy(
        cls,
        session_duration: Optional[float] = None,
        allow_multiwrite: Optional[bool] = False,
    ):
        """
        Convenience constructor to acquire a session directly from the proxy.
        Uses environment variables only and used for internal purposes.
        Parameters
        ----------
        session_duration : float
            Duration of the session in seconds.
            Default is 3600 seconds (1 hour).
        allow_multiwrite : bool
            Whether to allow other sessions to write to datasource
            already being written to by this session.
            Default is False.
        """

        try:
            res = retried_request(
                f"{os.environ['DATAMESH_ZARR_PROXY']}/session/",
                params={
                    "duration": session_duration or 3600,
                    "allow_multiwrite": allow_multiwrite,
                },
                headers={
                    "X-DATAMESH-TOKEN": os.environ["DATAMESH_TOKEN"],
                    "USER": os.environ["DATAMESH_USER"],
                    "Cache-Control": "no-cache",
                },
            )
            if res.status_code != 200:
                raise DatameshConnectError(
                    "Failed to create session with error: " + res.text
                )
            session = cls(**res.json())
            session._connection = lambda: None
            session._connection._gateway = os.environ["DATAMESH_ZARR_PROXY"]
            session._connection._is_v1 = True
            atexit.register(session.close)
            return session
        except Exception as e:
            raise DatameshSessionError(
                f"Error when acquiring datamesh session from proxy {e}"
            )

    @classmethod
    def from_session_id(cls, connection, session_id: str):
        """
        Acquire a session from the connection using the session id.

        Parameters
        ----------
        connection : Connection
            Connection object to acquire session from.
        session_id : str
            Session id to acquire.
        """

        # Back-compatibility with beta version (returning dummy session object)
        if not connection._is_v1:
            raise DatameshSessionError(
                "Cannot acquire session from id when using datamesh v0"
            )
        # v1
        try:
            res = retried_request(
                f"{connection._gateway}/session/{session_id}",
                headers=connection._auth_headers,
            )
            if res.status_code != 200:
                raise DatameshConnectError(
                    f"Failed to retrieve session {session_id} with error: {res.text}"
                )
            session = cls(**res.json())
            session._connection = connection
            return session
        except Exception as e:
            raise DatameshSessionError(f"Error when acquiring datamesh session {e}")

    @property
    def header(self):
        return {"X-DATAMESH-SESSIONID": self.id}

    def add_header(self, headers: dict):
        return {**headers, **self.header}

    def close(self, finalise_write: bool = False):
        # Back-compatibility with beta version (ignoring)
        if not self._connection._is_v1:
            return
        # datamesh v1
        try:
            atexit.unregister(self.close)
        except:
            pass
        res = retried_request(
            f"{self._connection._gateway}/session/{self.id}",
            method="DELETE",
            params={"finalise_write": finalise_write},
            headers=self.header,
        )
        if res.status_code != 204:
            if finalise_write:
                raise DatameshConnectError(
                    "Failed to finalise write with error: " + res.text
                )
            print("Failed to close session with error: " + res.text)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # When using context manager, close the session
        # and finalise the write if no exception was raised
        self.close(finalise_write=exc_type is None)



import click

key = click.style(u"\U0001F511", fg='yellow')

watch = click.style("\u23F1", fg='white')

globe = click.style('\U0001F30D', fg='blue')

spin = click.style(u"\u21BB", fg='cyan')

err = click.style('\u2715', fg='red')

chk = click.style('\u2713', fg='green')

wrn = click.style('\u26A0', fg='yellow')

info = click.style('\U0001F6C8', fg='cyan')


import os
import sys
from importlib.metadata import entry_points
from oceanum.cli import main

CLI_DEBUG = bool(os.getenv('OCEANUM_CLI_DEBUG'))

for run_ep in entry_points(group='oceanum.cli.extensions'):
    try:
        if CLI_DEBUG:
            print(f"Loading entry point: {run_ep.name}...")
        ep_module = run_ep.load()  
        if CLI_DEBUG:
            print(f"Imported module '{ep_module.__name__}' successfully.")
    except ModuleNotFoundError as e:
        print(f"Error loading entry point {run_ep.name}: {e}", file=sys.stderr)
        pass

if __name__ == "__main__":
    main()



# Create a base class for renderering the output of the command line interface.
# The outputs are rendered in the following formats:
# - Table
# - JSON
# - YAML
# The input format should be a dictionary or pydantic a model, 
# plus an indexing of headers for the table format.

from os import linesep
from typing import Type, Literal, Callable, Any
import json
import pprint

import yaml
import jsonpath
import click
from tabulate import tabulate

from pydantic import BaseModel

_sty = click.style

output_format_option = click.option(
    '-o','--output', 
    type=click.Choice(['table', 'json', 'yaml']), 
    default='table',
    help='Output format'
)

class RenderField(BaseModel):
    default: Any|None = None
    label: str = 'Name'
    path: str = '$.name'
    mod: Callable = lambda x: x
    lmod: Callable = lambda x: x
    sep: str = ', '

class Renderer:

    def __init__(self, 
        data:list|dict|Type[BaseModel], 
        fields: list[RenderField],
        indent: int = 0,
        output: Literal['table', 'json', 'yaml'] = 'table',
        ignore_fields: list[str] | None = None
    ) -> None:
        self.raw_data = data
        self.parsed_data = self._init_data(data)
        self.fields = fields
        self.indent = indent
        self.ignore_fields = ignore_fields or []

    def _init_data(self, data: list[dict]|list[Type[BaseModel]]|dict|Type[BaseModel]) -> list[dict]:
        """
        Convert the input data to a list of dictionaries.
        """
        dict_data = []
        if isinstance(data, list):
            for item in data:
                if isinstance(item, BaseModel):
                    dict_data.append(item.model_dump(mode='json'))
                else:
                    dict_data.append(item)
        elif isinstance(data, BaseModel):
            dict_data.append(data.model_dump(mode='json'))
        else:
            dict_data.append(data)
        return dict_data
        
    def render_table(self, tablefmt='simple', **dump_kwargs) -> str:
        table_data = []
        headers = [f.label for f in self.fields]
        indent = self.indent if isinstance(self.indent,int) else 0
        for item in self.parsed_data:
            row = []
            for field in self.fields:
                matches = jsonpath.findall(field.path, item)
                if matches:
                    row.append(field.sep.join(str(field.mod(m)) for m in field.lmod(matches)))
                else:
                    row.append(None)
                    click.echo(f"{_sty('WARNING', fg='yellow')}: Could not find a data field for '{field.label}' at path '{field.path}'")
            table_data.append(row)
        if tablefmt == 'plain':
            table_data = zip(headers, table_data[0])
            return tabulate(table_data, tablefmt=tablefmt, **dump_kwargs)
        else:
            return tabulate(table_data, headers=headers, tablefmt=tablefmt)

    def render_json(self, **dump_kwargs) -> str:
        return json.dumps(self.parsed_data, **dump_kwargs)

    def render_yaml(self, **dump_kwargs) -> str:
        return yaml.dump(self.parsed_data, **dump_kwargs)
    
    def render(self, output_format:str='plain', **dump_kwargs) -> str:
        output = ''
        if output_format == 'table':
            output = self.render_table(**dump_kwargs)
        elif output_format == 'json':
            output = self.render_json(**dump_kwargs)
        elif output_format == 'yaml':
            output = self.render_yaml(**dump_kwargs)
        else:
            raise ValueError(f"Invalid format: {output_format}")
        return linesep.join([' '*self.indent+l for l in output.split(linesep)])


from unittest.mock import patch

from click.testing import CliRunner

from oceanum.cli import auth
from oceanum.cli import models

def test_auth_help():
    result = CliRunner().invoke(auth.auth, ['--help'])
    assert result.exit_code == 0

def test_auth_login_help():
    result = CliRunner().invoke(auth.login, ['--help'])
    assert result.exit_code == 0

def test_auth_logout_help():
    result = CliRunner().invoke(auth.logout, ['--help'])
    assert result.exit_code == 0

def test_auth_login():
    token = models.TokenResponse(
        access_token='123',
        expires_in=3600,
        token_type='Bearer',
        domain='oceanum.test'
    )
    with patch.object(auth.Auth0Client, 'get_token') as mock_save:
        mock_save.return_value = token
        result = CliRunner().invoke(auth.main, ['auth', 'login'])
    assert result.exit_code == 0



import click

from ..storage import filesystem

from . import main
from .utils import bytes_to_human, item_to_long, get_bearer_token
from .auth import login_required

# =====================================================================================
# Storage CLI
# =====================================================================================
@main.group()
def storage():
    """Oceanum Storage commands"""
    click.help_option('-h', '--help')

@storage.command()
@click.option("-l", "--long", is_flag=True, help="Long listing format")
@click.option("-h", "--human-readable", is_flag=True, help="Readable sizes with -l")
@click.option("-r", "--recursive", is_flag=True, help="List subdirectories recursively")
@click.argument("path", default="/")
@click.pass_context
@login_required
def ls(
    ctx: click.Context,
    path: str,
    long: bool,
    human_readable: bool,
    recursive: bool,
):
    """List contents in the oceanum storage (the root directory by default)."""
    contents = filesystem.ls(
        path=path,
        recursive=recursive,
        detail=long,
        token=get_bearer_token(ctx),
        service=f"https://storage.{ctx.obj.domain}/",
    )
    if long:
        sizes = 0
        for item in contents.values(): # type: ignore
            line, size = item_to_long(item, human_readable=human_readable)
            sizes += size
            click.echo(line)
        click.echo(
            f"TOTAL: {len(contents)} objects, {sizes} bytes ({bytes_to_human(sizes)})"
        )
    else:
        click.echo("\n".join(contents))


@storage.command()
@click.option("-r", "--recursive", is_flag=True, help="Copy directories recursively")
@click.argument("source")
@click.argument("dest")
@click.pass_context
def get(
    ctx: click.Context,
    recursive: bool,
    source: str,
    dest: str,
):
    """Copy content from SOURCE to DEST."""
    filesystem.get(
        source=source,
        dest=dest,
        recursive=recursive,
        token=get_bearer_token(ctx),
        service=f"https://storage.{ctx.obj.domain}/",
    )


@storage.command()
@click.option("-r", "--recursive", is_flag=True, help="Copy directories recursively")
@click.argument("source")
@click.argument("dest")
@click.pass_context
#
def put(
    ctx: click.Context,
    recursive: bool,
    source: str,
    dest: str,
):
    """Copy content from SOURCE to DEST."""
    filesystem.put(
        source=source,
        dest=dest,
        recursive=recursive,
        token=get_bearer_token(ctx),
        service=f"https://storage.{ctx.obj.domain}/",
    )


@storage.command()
@click.option("-r", "--recursive", is_flag=True, help="Remove directories recursively")
@click.option("-f", "--force", is_flag=True, help="Force removal without confirmation")
@click.argument("path")
@click.pass_context
@login_required
def rm(ctx: click.Context, recursive: bool, force: bool, path: str):
    """Remove PATH."""
    if not force:
        if recursive:
            message = f"Are you sure you want to recursively remove '{path}' and all its contents?"
        else:
            message = f"Are you sure you want to remove '{path}'?"

        if not click.confirm(message):
            click.echo("Operation cancelled.")
            return

    try:
        filesystem.rm(
            path=path,
            recursive=recursive,
            token=get_bearer_token(ctx),
            service=f"https://storage.{ctx.obj.domain}/",
        )
        click.echo(f"Successfully removed: {path}")
    except FileNotFoundError:
        click.echo(f"Error: Path '{path}' not found", err=True)
        ctx.exit(1)
    except OSError as e:
        if "Directory not empty" in str(e):
            click.echo(f"Error: Directory '{path}' is not empty. Use -r/--recursive to remove non-empty directories.", err=True)
        else:
            click.echo(f"Error: {e}", err=True)
        ctx.exit(1)
    except Exception as e:
        click.echo(f"Error removing '{path}': {e}", err=True)
        ctx.exit(1)


@storage.command()
@click.argument("path")
@click.pass_context
@login_required
def exists(ctx: click.Context, path: str):
    """Check if PATH exists in storage."""
    try:
        result = filesystem.exists(
            path=path,
            token=get_bearer_token(ctx),
            service=f"https://storage.{ctx.obj.domain}/",
        )
        if result:
            click.echo(f"EXISTS: {path}")
        else:
            click.echo(f"NOT FOUND: {path}")
            ctx.exit(1)
    except Exception as e:
        click.echo(f"Error checking '{path}': {e}", err=True)
        ctx.exit(1)


@storage.command()
@click.argument("path")
@click.pass_context
@login_required
def isfile(ctx: click.Context, path: str):
    """Check if PATH is a file in storage."""
    try:
        result = filesystem.isfile(
            path=path,
            token=get_bearer_token(ctx),
            service=f"https://storage.{ctx.obj.domain}/",
        )
        if result:
            click.echo(f"FILE: {path}")
        else:
            click.echo(f"NOT A FILE: {path}")
            ctx.exit(1)
    except Exception as e:
        click.echo(f"Error checking '{path}': {e}", err=True)
        ctx.exit(1)


@storage.command()
@click.argument("path")
@click.pass_context
@login_required
def isdir(ctx: click.Context, path: str):
    """Check if PATH is a directory in storage."""
    try:
        result = filesystem.isdir(
            path=path,
            token=get_bearer_token(ctx),
            service=f"https://storage.{ctx.obj.domain}/",
        )
        if result:
            click.echo(f"DIRECTORY: {path}")
        else:
            click.echo(f"NOT A DIRECTORY: {path}")
            ctx.exit(1)
    except Exception as e:
        click.echo(f"Error checking '{path}': {e}", err=True)
        ctx.exit(1)
